/*
 * Zmanim Java API
 * Copyright (C) 2004-2007 Eliyahu Hershfeld
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc. 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA or connect to: http://www.fsf.org/copyleft/gpl.html
 */
package net.sourceforge.zmanim;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;
//import java.util.InvalidPropertiesFormatException;
import java.util.List;
//import java.util.TimeZone;
// import java.io.FileInputStream;
//import java.io.IOException;
import java.lang.reflect.Method;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import net.sourceforge.zmanim.util.AstronomicalCalculator;
import net.sourceforge.zmanim.util.NOAACalculator;
import net.sourceforge.zmanim.util.GeoLocation;
import net.sourceforge.zmanim.util.ZmanimFormatter;
import net.sourceforge.zmanim.util.Zman;

/**
 * A Java calendar that calculates astronomical time calculations such as
 * {@link #getSunrise() sunrise} and {@link #getSunset() sunset} times. This
 * class contains a {@link #getCalendar() Calendar} and can therefore use the
 * standard Calendar functionality to change dates etc. The calculation engine
 * used to calculate the astronomical times can be changed to a different
 * implementation by implementing the {@link AstronomicalCalculator} and setting
 * it with the {@link #setAstronomicalCalculator(AstronomicalCalculator)}. A
 * number of different implementations are included in the util package <br />
 * <b>Note:</b> There are times when the algorithms can't calculate proper
 * values for sunrise and sunset. This is usually caused by trying to calculate
 * times for areas either very far North or South, where sunrise / sunset never
 * happen on that date. This is common when calculating twilight with a deep dip
 * below the horizon for locations as south as London in the northern
 * hemisphere. When the calculations encounter this condition a null will be
 * returned when a <code>{@link java.util.Date}</code> is expected and {@link Double#NaN} will
 * be returned when a double is expected. The reason that <code>Exception</code>s
 * are not thrown in these cases is due to the fact that such problems are
 * expected and will often be encountered when a calendar for a full year is
 * being generated.
 * 
 * @author &copy; Eliyahu Hershfeld 2004 - 2007
 * @version 1.1
 */
public class AstronomicalCalendar /* extends GregorianCalendar */{
	private static final long serialVersionUID = 1;

	/**
	 * 90&deg; below the vertical. Used for certain calculations.<br />
	 * <b>Note </b>: it is important to note the distinction between this zenith
	 * and the {@link AstronomicalCalculator#adjustZenith adjusted zenith} used
	 * for some solar calculations. This 90 zenith is only used because some
	 * calculations in some subclasses are historically calculated as an offset
	 * in reference to 90.
	 */
	public static final double GEOMETRIC_ZENITH = 90;

	/**
	 * Default value for Sun's zenith and true rise/set Zenith (used in this
	 * class and subclasses) is the angle that the center of the Sun makes to a
	 * line perpendicular to the Earth's surface. If the Sun were a point and
	 * the Earth were without an atmosphere, true sunset and sunrise would
	 * correspond to a 90&deg; zenith. Because the Sun is not a point, and
	 * because the atmosphere refracts light, this 90&deg; zenith does not, in
	 * fact, correspond to true sunset or sunrise, instead the center of the
	 * Sun's disk must lie just below the horizon for the upper edge to be
	 * obscured. This means that a zenith of just above 90&deg; must be used.
	 * The Sun subtends an angle of 16 minutes of arc (this can be changed via
	 * the {@link #setSunRadius(double)} method , and atmospheric refraction
	 * accounts for 34 minutes or so (this can be changed via the
	 * {@link #setRefraction(double)} method), giving a total of 50 arcminutes.
	 * The total value for ZENITH is 90+(5/6) or 90.8333333&deg; for true
	 * sunrise/sunset.
	 */
	// public static double ZENITH = GEOMETRIC_ZENITH + 5.0 / 6.0;
	/** Sun's zenith at civil twilight (96&deg;). */
	public static final double CIVIL_ZENITH = 96;

	/** Sun's zenith at nautical twilight (102&deg;). */
	public static final double NAUTICAL_ZENITH = 102;

	/** Sun's zenith at astronomical twilight (108&deg;). */
	public static final double ASTRONOMICAL_ZENITH = 108;

	/** constant for milliseconds in a minute (60,000) */
	static final long MINUTE_MILLIS = 60 * 1000;

	private Calendar calendar;

	private GeoLocation geoLocation;

	private AstronomicalCalculator astronomicalCalculator;

	/**
	 * The getSunrise method Returns a <code>Date</code> representing the
	 * sunrise time. The zenith used for the calculation uses
	 * {@link #GEOMETRIC_ZENITH geometric zenith} of 90&deg;. This is adjusted by
	 * the {@link AstronomicalCalculator} that adds approximately 50/60 of a
	 * degree to account for 34 archminutes of refraction and 16 archminutes for
	 * the sun's radius for a total of
	 * {@link AstronomicalCalculator#adjustZenith 90.83333&deg;}. See documentation
	 * for the specific implementation of the {@link AstronomicalCalculator}
	 * that you are using.
	 * 
	 * @return the <code>Date</code> representing the exact sunrise time. If
	 *         the calculation can not be computed null will be returned.
	 * @see AstronomicalCalculator#adjustZenith
	 */
	public Date getSunrise() {
		double sunrise = getUTCSunrise(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunrise)) {
			return null;
		} else {
			sunrise = getOffsetTime(sunrise);
			return getDateFromTime(sunrise);
		}
	}

	/**
	 * Method that returns the sunrise without correction for elevation.
	 * Non-sunrise and sunset calculations such as dawn and dusk, depend on the
	 * amount of visible light, something that is not affected by elevation.
	 * This method returns sunrise calculated at sea level. This forms the base
	 * for dawn calculations that are calculated as a dip below the horizon
	 * before sunrise.
	 * 
	 * @return the <code>Date</code> representing the exact sea-level sunrise
	 *         time. If the calculation can not be computed null will be
	 *         returned.
	 * @see AstronomicalCalendar#getSunrise
	 * @see AstronomicalCalendar#getUTCSeaLevelSunrise
	 */
	public Date getSeaLevelSunrise() {
		double sunrise = getUTCSeaLevelSunrise(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunrise)) {
			return null;
		} else {
			sunrise = getOffsetTime(sunrise);
			return getDateFromTime(sunrise);
		}
	}

	/**
	 * A method to return the the beginning of civil twilight (dawn) using a
	 * zenith of {@link #CIVIL_ZENITH 96&deg;}.
	 * 
	 * @return The <code>Date</code> of the beginning of civil twilight using
	 *         a zenith of 96&deg;. If the calculation can not be computed null will
	 *         be returned.
	 * @see #CIVIL_ZENITH
	 */
	public Date getBeginCivilTwilight() {
		return getSunriseOffsetByDegrees(CIVIL_ZENITH);
	}

	/**
	 * A method to return the the beginning of nautical twilight using a zenith
	 * of {@link #NAUTICAL_ZENITH 102&deg;}.
	 * 
	 * @return The <code>Date</code> of the beginning of nautical twilight
	 *         using a zenith of 102&deg;. If the calculation can not be computed
	 *         null will be returned.
	 * @see #NAUTICAL_ZENITH
	 */
	public Date getBeginNauticalTwilight() {
		return getSunriseOffsetByDegrees(NAUTICAL_ZENITH);
	}

	/**
	 * A method that returns the the beginning of astronomical twilight using a
	 * zenith of {@link #ASTRONOMICAL_ZENITH 108&deg;}.
	 * 
	 * @return The <code>Date</code> of the beginning of astronomical twilight
	 *         using a zenith of 108&deg;. If the calculation can not be computed
	 *         null will be returned.
	 * @see #ASTRONOMICAL_ZENITH
	 */
	public Date getBeginAstronomicalTwilight() {
		return getSunriseOffsetByDegrees(ASTRONOMICAL_ZENITH);
	}

	/**
	 * The getSunset method Returns a <code>Date</code> representing the
	 * sunset time. The zenith used for the calculation uses
	 * {@link #GEOMETRIC_ZENITH geometric zenith} of 90&deg;. This is adjusted by
	 * the {@link AstronomicalCalculator} that adds approximately 50/60 of a
	 * degree to account for 34 archminutes of refraction and 16 archminutes for
	 * the sun's radius for a total of
	 * {@link AstronomicalCalculator#adjustZenith 90.83333&deg;}. See documentation
	 * for the specific implementation of the {@link AstronomicalCalculator}
	 * that you are using.
	 * Note: In certain cases the calculates sunset will occur before sunrise. This
	 * will typically happen when a timezone other than the local timezone is used
	 * (calculating Los Angeles sunset using a GMT timezone for example). In this case
	 * the sunset date will be incremented to the following date.
	 * 
	 * @return the <code>Date</code> representing the exact sunset time. If
	 *         the calculation can not be computed null will be returned. If the 
	 *         time calculation
	 * @see AstronomicalCalculator#adjustZenith
	 */
	public Date getSunset() {
		double sunset = getUTCSunset(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunset)) {
			return null;
		} else {
			sunset = getOffsetTime(sunset);
			return getAdjustedSunsetDate(getDateFromTime(sunset), getSunrise());
		}
	}
	
	/**
	 * A method that  will roll the sunset time forward a day if sunset occurs before
	 * sunrise. This will typically happen when a timezone other than the local timezone
	 * is used (calculating Los Angeles sunset using a GMT timezone for example).
	 * In this case the sunset date will be incremented to the following date.
	 * @param sunset the sunset date to adjust if needed
	 * @param sunrise the sunrise to compare to the sunset
	 * @return the adjusted sunset date
	 */
	private Date getAdjustedSunsetDate(Date sunset, Date sunrise){
		if(sunset != null && sunrise != null && sunrise.compareTo(sunset) >= 0){
			Calendar clonedCalendar = (GregorianCalendar)getCalendar().clone();
			clonedCalendar.setTime(sunset);
			clonedCalendar.add(Calendar.DAY_OF_MONTH, 1);
			return clonedCalendar.getTime();
		} else {
			return sunset;
		}
		
	}

	/**
	 * Method that returns the sunset without correction for elevation.
	 * Non-sunrise and sunset calculations such as dawn and dusk, depend on the
	 * amount of visible light, something that is not affected by elevation.
	 * This method returns sunset calculated at sea level. This forms the base
	 * for dusk calculations that are calculated as a dip below the horizon
	 * after sunset.
	 * 
	 * @return the <code>Date</code> representing the exact sea-level sunset
	 *         time. If the calculation can not be computed null will be
	 *         returned.
	 * @see AstronomicalCalendar#getSunset
	 * @see AstronomicalCalendar#getUTCSeaLevelSunset
	 */
	public Date getSeaLevelSunset() {
		double sunset = getUTCSeaLevelSunset(GEOMETRIC_ZENITH);
		if (Double.isNaN(sunset)) {
			return null;
		} else {
			sunset = getOffsetTime(sunset);
			//return getDateFromTime(sunset);
			return getAdjustedSunsetDate(getDateFromTime(sunset), getSeaLevelSunrise());
		}
	}

	/**
	 * A method to return the the end of civil twilight using a zenith of
	 * {@link #CIVIL_ZENITH 96&deg;}.
	 * 
	 * @return The <code>Date</code> of the end of civil twilight using a
	 *         zenith of {@link #CIVIL_ZENITH 96&deg;}. If the calculation can not
	 *         be computed null will be returned.
	 * @see #CIVIL_ZENITH
	 */
	public Date getEndCivilTwilight() {
		return getSunsetOffsetByDegrees(CIVIL_ZENITH);
	}

	/**
	 * A method to return the the end of nautical twilight using a zenith of
	 * {@link #NAUTICAL_ZENITH 102&deg;}.
	 * 
	 * @return The <code>Date</code> of the end of nautical twilight using a
	 *         zenith of {@link #NAUTICAL_ZENITH 102&deg;}. If the calculation can not be computed null will
	 *         be returned.
	 * @see #NAUTICAL_ZENITH
	 */
	public Date getEndNauticalTwilight() {
		return getSunsetOffsetByDegrees(NAUTICAL_ZENITH);
	}

	/**
	 * A method to return the the end of astronomical twilight using a zenith of
	 * {@link #ASTRONOMICAL_ZENITH 108&deg;}.
	 * 
	 * @return The The <code>Date</code> of the end of astronomical twilight
	 *         using a zenith of {@link #ASTRONOMICAL_ZENITH 108&deg;}. If the calculation can not be computed
	 *         null will be returned.
	 * @see #ASTRONOMICAL_ZENITH
	 */
	public Date getEndAstronomicalTwilight() {
		return getSunsetOffsetByDegrees(ASTRONOMICAL_ZENITH);
	}

	/**
	 * Utility method that returns a time offset. This method casts the offset
	 * as a <code>long</code> and calls {@link #getTimeOffset(Date, long)}.
	 * 
	 * @param time
	 *            the start time
	 * @param offset
	 *            the offset in milliseconds to add to the time
	 * @return the {@link java.util.Date}with the offset added to it
	 */
	public Date getTimeOffset(Date time, double offset) {
		return getTimeOffset(time, (long) offset);
	}

	/**
	 * A utility method to return a time offset.
	 * 
	 * @param time
	 *            the start time
	 * @param offset
	 *            the offset in milliseconds to add to the time.
	 * @return the {@link java.util.Date} with the offset in milliseconds added
	 *         to it
	 */
	public Date getTimeOffset(Date time, long offset) {
		if (time == null || offset == Long.MIN_VALUE) {
			return null;
		}
		return new Date(time.getTime() + offset);
	}

	/**
	 * A utility method to return the time of an offset by degrees below or
	 * above the horizon of {@link #getSunrise() sunrise}.
	 * 
	 * @param offsetZenith
	 *            the degrees before {@link #getSunrise()} to use in the
	 *            calculation. For time after sunrise use negative numbers.
	 * @return The {@link java.util.Date} of the offset after (or before)
	 *         {@link #getSunrise()}. If the calculation can not be computed
	 *         null will be returned.
	 */
	public Date getSunriseOffsetByDegrees(double offsetZenith) {
		double alos = getUTCSunrise(offsetZenith);
		if (Double.isNaN(alos)) {
			return null;
		} else {
			alos = getOffsetTime(alos);
			return getDateFromTime(alos);
		}
	}

	/**
	 * A utility method to return the time of an offset by degrees below or
	 * above the horizon of {@link #getSunset() sunset}.
	 * 
	 * @param offsetZenith
	 *            the degrees after {@link #getSunset()} to use in the
	 *            calculation. For time before sunset use negative numbers.
	 * @return The {@link java.util.Date}of the offset after (or before)
	 *         {@link #getSunset()}. If the calculation can not be computed
	 *         null will be returned.
	 */
	public Date getSunsetOffsetByDegrees(double offsetZenith) {
		double sunset = getUTCSunset(offsetZenith);
		if (Double.isNaN(sunset)) {
			return null;
		} else {
			sunset = getOffsetTime(sunset);
			//return getDateFromTime(sunset);
			return getAdjustedSunsetDate(getDateFromTime(sunset), getSunriseOffsetByDegrees(offsetZenith));
			
		}
	}

	/**
	 * Default constructor will set a default {@link GeoLocation}, use the
	 * default ({@link AstronomicalCalculator#getDefault()})
	 * AstronomicalCalculator and default the calendar to the current time. 
	 */
	public AstronomicalCalendar() {
		// setGeoLocation(new GeoLocation()); //PMD
		geoLocation = new GeoLocation();
		calendar = Calendar.getInstance(geoLocation.getTimeZone());
		astronomicalCalculator = AstronomicalCalculator.getDefault();
	}

	/**
	 * A constructor that takes in as a parameter geolocation information
	 * 
	 * @param geoLocation
	 *            The location information used for astronomical calculating sun
	 *            times.
	 */
	public AstronomicalCalendar(GeoLocation geoLocation) {
		this(); // call default constructor to initialize astronomicalGeometryCalculator
		setGeoLocation(geoLocation);
		calendar = Calendar.getInstance(geoLocation.getTimeZone());
	}

	/**
	 * Method that returns the sunrise in UTC time without correction for time
	 * zone offset from GMT and without using daylight savings time.
	 * 
	 * @param zenith
	 *            the degrees below the horizon. For time after sunrise use
	 *            negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 */
	public double getUTCSunrise(double zenith) {
		return getAstronomicalCalculator().getUTCSunrise(this, zenith, true);
	}

	/**
	 * Method that returns the sunrise in UTC time without correction for time
	 * zone offset from GMT and without using daylight savings time. Non-sunrise
	 * and sunset calculations such as dawn and dusk, depend on the amount of
	 * visible light, something that is not affected by elevation. This method
	 * returns UTC sunrise calculated at sea level. This forms the base for dawn
	 * calculations that are calculated as a dip below the horizon before
	 * sunrise.
	 * 
	 * @param zenith
	 *            the degrees below the horizon. For time after sunrise use
	 *            negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 * @see AstronomicalCalendar#getUTCSunrise
	 * @see AstronomicalCalendar#getUTCSeaLevelSunset
	 */
	public double getUTCSeaLevelSunrise(double zenith) {
		return getAstronomicalCalculator().getUTCSunrise(this, zenith, false);
	}

	/**
	 * Method that returns the sunset in UTC time without correction for time
	 * zone offset from GMT and without using daylight savings time.
	 * 
	 * @param zenith
	 *            the degrees below the horizon. For time after before sunset
	 *            use negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 * @see AstronomicalCalendar#getUTCSeaLevelSunset
	 */
	public double getUTCSunset(double zenith) {
		return getAstronomicalCalculator().getUTCSunset(this, zenith, true);
	}

	/**
	 * Method that returns the sunset in UTC time without correction for
	 * elevation, time zone offset from GMT and without using daylight savings
	 * time. Non-sunrise and sunset calculations such as dawn and dusk, depend
	 * on the amount of visible light, something that is not affected by
	 * elevation. This method returns UTC sunset calculated at sea level. This
	 * forms the base for dusk calculations that are calculated as a dip below
	 * the horizon after sunset.
	 * 
	 * @param zenith
	 *            the degrees below the horizon. For time before sunset use
	 *            negative numbers.
	 * @return The time in the format: 18.75 for 18:45:00 UTC/GMT. If the
	 *         calculation can not be computed {@link Double#NaN} will be
	 *         returned.
	 * @see AstronomicalCalendar#getUTCSunset
	 * @see AstronomicalCalendar#getUTCSeaLevelSunrise
	 */
	public double getUTCSeaLevelSunset(double zenith) {
		return getAstronomicalCalculator().getUTCSunset(this, zenith, false);
	}

	/**
	 * A method that adds time zone offset and daylight savings time to the raw
	 * UTC time.
	 * 
	 * @param time
	 *            The UTC time to be adjusted.
	 * @return The time adjusted for the time zone offset and daylight savings
	 *         time.
	 */
	private double getOffsetTime(double time) {
		boolean dst = getCalendar().getTimeZone().inDaylightTime(
				getCalendar().getTime());
		double dstOffset = 0;
		// be nice to Newfies and use a double
		double gmtOffset = getCalendar().getTimeZone().getRawOffset()
				/ (60 * MINUTE_MILLIS);
		if (dst) {
			dstOffset = getCalendar().getTimeZone().getDSTSavings()
					/ (60 * MINUTE_MILLIS);
		}
		return time + gmtOffset + dstOffset;
	}

	/**
	 * Method to return a temporal (solar) hour. The day from sunrise to sunset
	 * is split into 12 equal parts with each one being a temporal hour.
	 * 
	 * @return the <code>long</code> millisecond length of a temporal hour. If
	 *         the calculation can not be computed {@link Long#MIN_VALUE} will
	 *         be returned.
	 */
	public long getTemporalHour() {
		return getTemporalHour(getSunrise(), getSunset());
	}

	/**
	 * Utility method that will allow the calculation of a temporal (solar) hour
	 * based on the sunrise and sunset passed to this method.
	 * 
	 * @param sunrise
	 *            The start of the day.
	 * @param sunset
	 *            The end of the day.
	 * @see #getTemporalHour()
	 * @return the <code>long</code> millisecond length of the temporal hour.
	 *         If the calculation can not be computed {@link Long#MIN_VALUE}
	 *         will be returned.
	 */
	public long getTemporalHour(Date sunrise, Date sunset) {
		if (sunrise == null || sunset == null) {
			return Long.MIN_VALUE;
		}
		return (sunset.getTime() - sunrise.getTime()) / 12;
	}

	/**
	 * method that returns sundial noon. This is calculated as halfway between
	 * sunrise and sunset.
	 * 
	 * @return the <code>Date</code> representing Sun's transit. If the
	 *         calculation can not be computed null will be returned.
	 */
	public Date getSunTransit() {
		return getTimeOffset(getSunrise(), getTemporalHour() * 6);
	}

	/**
	 * A method that returns a <code>Date</code> from the time passed in
	 * 
	 * @param time
	 *            The time to be set as the time for the <code>Date</code>.
	 *            The time expected is in the format: 18.75 for 6:45:00 PM
	 * @return The Date.
	 */
	private Date getDateFromTime(double time) {
		//System.out.println(time);
		if (Double.isNaN(time)) {
			return null;
		}
		Calendar cal = new GregorianCalendar();
		cal.clear();
		cal.set(Calendar.YEAR, getCalendar().get(Calendar.YEAR));
		cal.set(Calendar.MONTH, getCalendar().get(Calendar.MONTH));
		cal.set(Calendar.DAY_OF_MONTH, getCalendar().get(Calendar.DAY_OF_MONTH));
		// sunset time (usually offset) is later than midnight GMT
		//FIXME, with the use of getCorrectedSunsetDate(), the following might not be needed
		if ((!getCalendar().getTimeZone().inDaylightTime(getCalendar().getTime()) && time < 0)
				|| (getCalendar().getTimeZone().inDaylightTime(getCalendar().getTime()) && time < 1)) {
			cal.add(Calendar.DAY_OF_MONTH, 1);
		}

		int hours = (int) time; // cut off minutes
		boolean prob = false;
		if (hours < 1) {
			prob = true;
		}
		time -= hours;

		int minutes = (int) (time *= 60);
		time -= minutes;
		int seconds = (int) (time *= 60);
		time -= seconds; // milliseconds

		cal.set(Calendar.HOUR_OF_DAY, hours);
		cal.set(Calendar.MINUTE, minutes);
		cal.set(Calendar.SECOND, seconds);
		cal.set(Calendar.MILLISECOND, (int) (time * 1000));
		if (prob == true) {//FIXME: might not be an issue with the new getCorrectedSunsetDate
			// System.out.println(cal);
		}
		return cal.getTime();
	}

	/**
	 * @return an XML formatted representation of the class. It returns the
	 *         default output of the {@link #toXML() toXML}
	 *         method.
	 * @see #toXML()
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return toXML();
	}

	/**
	 * @see java.lang.Object#equals(Object)
	 */
	public boolean equals(Object object) {
		if (this == object)
			return true;
		if (!(object instanceof AstronomicalCalendar))
			return false;
		AstronomicalCalendar aCal = (AstronomicalCalendar) object;
		return getCalendar().equals(aCal.getCalendar())
				&& getGeoLocation().equals(aCal.getGeoLocation())
				&& getAstronomicalCalculator().equals(
						aCal.getAstronomicalCalculator());
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		int result = 17;
		// needed or this and subclasses will return identical hash
		result = 37 * result + getClass().hashCode();
		result += 37 * result + getCalendar().hashCode();
		result += 37 * result + getGeoLocation().hashCode();
		result += 37 * result + getAstronomicalCalculator().hashCode();
		return result;
	}

	/**
	 * @return Returns the geoLocation.
	 */
	public GeoLocation getGeoLocation() {
		return geoLocation;
	}

	/**
	 * @param geoLocation
	 *            The geoLocation to set.
	 */
	public void setGeoLocation(GeoLocation geoLocation) {
		this.geoLocation = geoLocation;
		getCalendar().setTimeZone(geoLocation.getTimeZone());// TODO might
		// not be needed
	}

	/**
	 * A method to return the current AstronomicalCalculator set.
	 * 
	 * @return Returns the astronimicalCalculator.
	 * @see #setAstronomicalCalculator(AstronomicalCalculator)
	 */
	public AstronomicalCalculator getAstronomicalCalculator() {
		return astronomicalCalculator;
	}

	/**
	 * A method to set the {@link AstronomicalCalculator} used for astronomical
	 * calculations. The Zmanim package ships with a number of different
	 * implementations of the <code>abstract</code>
	 * {@link AstronomicalCalculator} based on different algorithms, including
	 * {@link net.sourceforge.zmanim.util.SunTimesCalculator one implementation}
	 * based on the <a href = "http://aa.usno.navy.mil/">US Naval Observatory's</a>
	 * algorithm, and
	 * {@link net.sourceforge.zmanim.util.JSuntimeCalculator another} based on
	 * <a href=""http://noaa.gov">NOAA's</a> algorithm. This allows easy
	 * runtime switching and comparison of different algorithms.
	 * 
	 * @param astronomicalCalculator
	 *            The astronimicalCalculator to set.
	 */
	public void setAstronomicalCalculator(
			AstronomicalCalculator astronomicalCalculator) {
		this.astronomicalCalculator = astronomicalCalculator;
	}

	/**
	 * returns the Calendar object encapsulated by this class.
	 * @return Returns the calendar.
	 */
	public Calendar getCalendar() {
		return calendar;
	}

	/**
	 * @param calendar
	 *            The calendar to set.
	 */
	public void setCalendar(Calendar calendar) {
		this.calendar = calendar;
	}

	/**
	 * A method that returns an XML formatted <code>String</code> representing
	 * the serialized <code>Object</code>.
	 * The format used is:
	 * 
	 * <pre>
	 *  &lt;AstronomicalTimes date=&quot;1969-02-08&quot; type=&quot;net.sourceforge.zmanim.AstronomicalCalendar algorithm=&quot;US Naval Almanac Algorithm&quot; location=&quot;Lakewood, NJ&quot; latitude=&quot;40.095965&quot; longitude=&quot;-74.22213&quot; elevation=&quot;31.0&quot; timeZoneName=&quot;Eastern Standard Time&quot; timeZoneID=&quot;America/New_York&quot; timeZoneOffset=&quot;-5&quot;&gt;
	 *     &lt;Sunrise&gt;2007-02-18T06:45:27-05:00&lt;/Sunrise&gt;
	 *     &lt;TemporalHour&gt;PT54M17.529S&lt;/TemporalHour&gt;
	 *     ...
	 *   &lt;/AstronomicalTimes&gt;   
	 * </pre>
	 * 
	 * Note that the output uses the <a
	 * href="http://www.w3.org/TR/xmlschema11-2/#dateTime">xsd:dateTime</a>
	 * format for times such as sunrise, and <a
	 * href="http://www.w3.org/TR/xmlschema11-2/#duration">xsd:duration</a>
	 * format for times that are a duration such as the length of a
	 * {@link #getTemporalHour() temporal hour}. The output of this
	 * method is returned by the {@link #toString() toString} }.
	 * 
	 * @return The XML formatted <code>String</code>. The format will be:
	 * 
	 * <pre>
	 *  &lt;AstronomicalTimes date=&quot;1969-02-08&quot; type=&quot;net.sourceforge.zmanim.AstronomicalCalendar algorithm=&quot;US Naval Almanac Algorithm&quot; location=&quot;Lakewood, NJ&quot; latitude=&quot;40.095965&quot; longitude=&quot;-74.22213&quot; elevation=&quot;31.0&quot; timeZoneName=&quot;Eastern Standard Time&quot; timeZoneID=&quot;America/New_York&quot; timeZoneOffset=&quot;-5&quot;&gt;
	 *     &lt;Sunrise&gt;2007-02-18T06:45:27-05:00&lt;/Sunrise&gt;
	 *     &lt;TemporalHour&gt;PT54M17.529S&lt;/TemporalHour&gt;
	 *     ...
	 *  &lt;/AstronomicalTimes&gt;   
	 * </pre>
	 * 
	 */
	public String toXML() {
		ZmanimFormatter formatter = new ZmanimFormatter(ZmanimFormatter.XSD_DURATION_FORMAT, new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss"));
		DateFormat df = new SimpleDateFormat("yyyy-MM-dd");
		String output = "<";
		if(getClass().getName().endsWith("AstronomicalCalendar")){
			output += "AstronomicalTimes";
		} else if(getClass().getName().endsWith("ZmanimCalendar")){
			output += "Zmanim";
		}
		output += " date=\"" + df.format(this.getCalendar().getTime()) + "\"";
		output += " type=\"" + getClass().getName() + "\"";
		output += " algorithm=\""
				+ getAstronomicalCalculator().getCalculatorName() + "\"";
		output += " location=\"" + getGeoLocation().getLocationName() + "\"";
		output += " latitude=\"" + getGeoLocation().getLatitude() + "\"";
		output += " longitude=\"" + getGeoLocation().getLongitude() + "\"";
		output += " elevation=\"" + getGeoLocation().getElevation() + "\"";
		output += " timeZoneName=\""
				+ getGeoLocation().getTimeZone().getDisplayName() + "\"";
		output += " timeZoneID=\"" + getGeoLocation().getTimeZone().getID()
				+ "\"";
		output += " timeZoneOffset=\""
				+ (getGeoLocation().getTimeZone().getOffset(
						getCalendar().getTimeInMillis()) / (60 * 60 * 1000))
				+ "\"";

		output += ">\n";

		Method[] theMethods = getClass().getMethods();
		String tagName = "";
		Object value = null;
		List dateList = new ArrayList();
		List durationList = new ArrayList();
		for (int i = 0; i < theMethods.length; i++) {
			if (includeMethod(theMethods[i])) {
				tagName = theMethods[i].getName().substring(3);
				try {
					value = theMethods[i].invoke(this, (Object[]) null);
					if (value instanceof Date) {
						
						dateList.add(new Zman((Date) value, tagName));
//						output += "\t<" + tagName;
//						output += ">";
//						output += formatter.formatDateTime((Date) value,
//								getCalendar())
//								+ "</" + tagName + ">\n";
					} else if (value instanceof Long) {// shaah zmanis
//						output += "\t<" + tagName;
//						output += ">"
//								+ formatter.format((int) ((Long) value)
//										.longValue()) + "</" + tagName + ">\n";
						durationList.add(new Zman((int) ((Long) value).longValue(), tagName));
					} else { // will probably never enter this block
						output += "<" + tagName + ">" + value + "</" + tagName
								+ ">\n";
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		Zman zman;
	 	Collections.sort(dateList, Zman.DATE_ORDER);
		for(int i=0; i< dateList.size(); i++){
			zman = (Zman)dateList.get(i);
			output += "\t<" + zman.getZmanLabel();
			output += ">";
			output += formatter.formatDateTime(zman.getZman(), getCalendar())
					+ "</" + zman.getZmanLabel() + ">\n";
		}
		for(int i=0; i< durationList.size(); i++){
			zman = (Zman)durationList.get(i);
			output += "\t<" + zman.getZmanLabel();
			output += ">";
			output += formatter.format((int) zman.getDuration())
					+ "</" + zman.getZmanLabel() + ">\n";
		}

		if(getClass().getName().endsWith("AstronomicalCalendar")){
			output += "</AstronomicalTimes>";
		} else if(getClass().getName().endsWith("ZmanimCalendar")){
			output += "</Zmanim>";
		}
		return output;
	}
	
	public String toXML2() {
		ZmanimFormatter formatter = new ZmanimFormatter(ZmanimFormatter.XSD_DURATION_FORMAT, new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss"));
		DateFormat df = new SimpleDateFormat("yyyy-MM-dd");
		String output = "<";
		if(getClass().getName().endsWith("AstronomicalCalendar")){
			output += "AstronomicalTimes";
		} else if(getClass().getName().endsWith("ZmanimCalendar")){
			output += "Zmanim";
		}
		output += " date=\"" + df.format(this.getCalendar().getTime()) + "\"";
		output += " type=\"" + getClass().getName() + "\"";
		output += " algorithm=\""
				+ getAstronomicalCalculator().getCalculatorName() + "\"";
		output += " location=\"" + getGeoLocation().getLocationName() + "\"";
		output += " latitude=\"" + getGeoLocation().getLatitude() + "\"";
		output += " longitude=\"" + getGeoLocation().getLongitude() + "\"";
		output += " elevation=\"" + getGeoLocation().getElevation() + "\"";
		output += " timeZoneName=\""
				+ getGeoLocation().getTimeZone().getDisplayName() + "\"";
		output += " timeZoneID=\"" + getGeoLocation().getTimeZone().getID()
				+ "\"";
		output += " timeZoneOffset=\""
				+ (getGeoLocation().getTimeZone().getOffset(
						getCalendar().getTimeInMillis()) / (60 * 60 * 1000))
				+ "\"";

		output += ">\n";

		Method[] theMethods = getClass().getMethods();
		String tagName = "";
		Object value = null;
		for (int i = 0; i < theMethods.length; i++) {
			if (includeMethod(theMethods[i])) {
				tagName = theMethods[i].getName().substring(3);
				try {
					value = theMethods[i].invoke(this, (Object[]) null);
					if (value instanceof Date) {
						output += "\t<" + tagName;
						// if(props!= null){
						// String zmanimDescription =
						// props.getProperty(tagName);
						// if(zmanimDescription != null){
						// output += " description=\"" + zmanimDescription +
						// "\"";
						// }
						// }
						output += ">";
						output += formatter.formatDateTime((Date) value,
								getCalendar())
								+ "</" + tagName + ">\n";
					} else if (value instanceof Long) {// shaah zmanis
						output += "\t<" + tagName;
						// if(props!= null){
						// String zmanimDescription =
						// props.getProperty(tagName);
						// if(zmanimDescription != null){
						// output += " description=\"" + zmanimDescription +
						// "\"";
						// }
						// }
						output += ">"
								+ formatter.format((int) ((Long) value)
										.longValue()) + "</" + tagName + ">\n";
					} else { // will probably never enter this block
						output += "<" + tagName + ">" + value + "</" + tagName
								+ ">\n";
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		if(getClass().getName().endsWith("AstronomicalCalendar")){
			output += "</AstronomicalTimes>";
		} else if(getClass().getName().endsWith("ZmanimCalendar")){
			output += "</Zmanim>";
		}
		return output;
	}

	private static boolean includeMethod(Method method) {
		List methodWhiteList = new ArrayList();
		// methodWhiteList.add("getName");

		List methodBlackList = new ArrayList();
		// methodBlackList.add("getGregorianChange");

		if (methodWhiteList.contains(method.getName()))
			return true;
		if (methodBlackList.contains(method.getName()))
			return false;

		if (method.getParameterTypes().length > 0)
			return false; // skip get methods with parameters
		if (!method.getName().startsWith("get"))
			return false;

		if (method.getReturnType().getName().endsWith("Date")
				|| method.getReturnType().getName().endsWith("long")) {
			return true;
		}
		return false;
	}
	
	public static void main(String [] args){
		
		String locationName = "Lakewood, NJ";
		// locationName = "Gateshead, United Kingdom";
		// locationName = "Whippany, NJ";
		// locationName = "Thule";
		double latitude = 40.095965; // Lakewood, NJ
		double longitude = -74.222130; // Lakewood, NJ
		
		//latitude = 34.052187; //LA, CA
		//longitude = -118.243425;// LA, CA 
		// latitude = 54.98333333; //Gateshead, United Kingdom
		// longitude = -1.583333333; //Gateshead, United Kingdom
		// latitude = 40.82444; //Whippany, NJ
		// longitude = -74.4175; //Whippany, NJ
		 //latitude = 76.32; //thule
		// longitude = 68.5;
		TimeZone timeZone = TimeZone.getTimeZone("America/New_York");
		// timeZone = TimeZone.getTimeZone("Europe/London");
		 //timeZone = TimeZone.getTimeZone("America/Thule");
		//timeZone = TimeZone.getTimeZone("GMT");

		GeoLocation location = new GeoLocation(locationName, latitude, longitude,
				timeZone);
		AstronomicalCalendar ac = new AstronomicalCalendar(location);
		ac.setAstronomicalCalculator(new NOAACalculator());
		ac.getCalendar().set(Calendar.YEAR, 2000);
		//System.out.println(ac);
		
		Date sunrise = ac.getSunrise();
		//Date sunset = ac.getCorrectedSunsetDate(sunrise, sunrise);
		
		System.out.println(sunrise);
		Date other = ac.getBeginAstronomicalTwilight();
		System.out.println(other);
	}
}